;;; monk-cdaudio.el --- routines for cd audio
;; Sun Dec 31 21:36:53 2000 split from monk.el

(eval-when-compile
  (require 'dired-aux)
  (require 'dired)
  (require 'cl))

(provide 'monk-cdaudio)

;;; Sun May 25 03:03:00 2003
;;; cdaudio variables moved from monk.el 

(defvar monk-player-command "cdda2wav"
  "*cdda2wav\(1\) player command.  Defaults \"cdda2wav\".
Possible alternative is:

\"nice -n 7 cdda2wav\"

Sometimes nice\(1\) might be required to lower priority to avoid
malfunction \(eg. X mouse pointer goes insane\).

See also docstring of `monk-command-options', `monk-post-filter'
etc., and `monk-generate-command'.")

(defvar monk-command-options '("-w")
  "*cdda2wav\(1\)'s option.  Defaults '\(\"-w\"\).  Other possible value:

'\(\"-w\" \"-S4\"\)

Options like \"-N\", \"-e\" etc. are automatically generated, depending
upon the value of `monk-post-filter' so you don't
have to care about them.  See also `monk-player-command'.")

(defvar monk-post-filter nil
  "*Appended to `player' command line for `start-process-shell-command'.
Defaults nil, and this nil value causes `monk-generate-command' to
add \"-e\" option to the command line, otherwise to remove the \"-e\"
option.

Possible value is:

\"bag | sox -t .wav - -t ossdsp /dev/dsp\"

which can be used to filter output of `monk-player-command' to /dev/dsp
in case the automatically generated \"-e\" won't work as expected.
Must be in string form, and \"|\" should not be prepended at the top.")

(defvar monk-cdda2wav-device (getenv "CDDA_DEVICE")
  "*Must setup correctly, or monk won't work.
If environment variable CDDA_DEVICE is set, its value is used.")

(defun monk-get-device-opt ()
  "The name tells all about it."
  (list
   (format "-D%s"
	   (or monk-cdda2wav-device
	       (error
		"You should setup `monk-cdda2wav-device' correctly.")))))

;;;
;;; Cd Info Ripper
;;;

(defvar monk-disc-info nil
  "Current disc info derived via cdda2wav\(1\) in very raw form.
Should not be used while operation.  Use `monk-track-list' instead.
No more than a failsafe.")

(defvar monk-track-list nil "Track list of current disc.")

(defsubst monk-get-current-line-list (&optional delim-regex)
  ;; Requires dired-split() -- but not require()'d. `dired' is always there.
  (dired-split (or delim-regex "[ \t]+")
	       (buffer-substring
		(save-excursion (beginning-of-line) (point))
		(save-excursion (end-of-line) (point)))))

(defvar monk-cdda2wav-query-command "cdda2wav -gNJ -v34")

(defun monk-get-info ()
  "Run `monk-cdda2wav-query-command' to get cd info.
Set up `monk-disc-info'and `monk-track-list'.
Parsing `monk-info-buffer' done by `monk-parse-info'."
  (let ((buf (get-buffer-create monk-info-buffer)) ret)
    (save-excursion
      (setq default-directory "~/")
      (set-buffer buf) (erase-buffer)
      (setq ret (monk-with-coding-system
		 (call-process shell-file-name nil t nil
			       "-c" monk-cdda2wav-query-command)))
      (if (not (= ret 0))
	  (error "`monk-get-info' failed."))
      (setq monk-disc-info (monk-parse-info)
	    monk-track-list (monk-get-track-list monk-disc-info))
      monk-disc-info)))

(defun monk-get-track-list (infolist)
  "Parse cdda2wav INFOLIST extract track-list to be bound to \
`monk-track-list' and return it."
  (let ((tracklist (nth 5 infolist))
	(tracks (let ((tem (car (car infolist))))
		  (substring tem (string-match "[0-9]+$" tem))))
	(total (car (cdr (nth 0 infolist))))
	(discid (nth 1 infolist))
	(atitle  (nth 4 infolist))
	(leadout (nth 6 infolist))
	length nextbeg trk lis)
    (cond
     (tracklist
      (while tracklist
	(setq trk (car tracklist)
	      tracklist (cdr tracklist)) ; Advance here
	(cond
	 (tracklist
	  (setq nextbeg (nth 1 (car tracklist))
		length (- (string-to-int nextbeg)
			  (string-to-int (nth 1 trk)))))
	 (t   (setq nextbeg (nth 1 (nth 6 infolist))
		    length (- (string-to-int nextbeg)
			      (string-to-int (nth 1 trk))))))
	(setq lis
	      (cons
	       (list
		(nth 0 trk)
		(nth 2 trk)
		(string-to-int (nth 1 trk))
		length
		(mapconcat 'identity (nthcdr 8 trk) " ")
		) lis))	)
      ;; Add disc data
      ;; Tracks become sublist here.
      (setq lis (cons
		 (list (car leadout) (string-to-int (nth 1 leadout)))
		 (list (nreverse lis))))
      (mapcar (lambda (data)
		(setq lis (cons data lis)))
	      (list (list
		     (format "Tracks: %s" tracks)
		     (format "Total:  %s" total))
		    atitle discid))
      (nreverse lis))
     ;; Nothing will be inserted.
     (t nil))))

(defun monk-parse-info ()
  "Parse `monk-info-buffer' and return cd info as list."
  (save-excursion
    (set-buffer monk-info-buffer)
    (let (tracks
	  discid
	  cd-text
	  cd-extra
	  title
	  track-list leadout)
      ;;(widen)
      (goto-char (point-min))
      (and
       ;; Ugly & cheap. Depends on Line order...
       (search-forward-regexp "^Tracks:" nil t) (forward-line 0)
       (setq tracks (monk-get-current-line-list))

       (search-forward-regexp "^CDDB discid:" nil t) (forward-line 0)
       (setq discid (monk-get-current-line-list))

       (search-forward-regexp "^CD-Text:" nil t) (forward-line 0)
       (setq cd-text (monk-get-current-line-list))

       (search-forward-regexp "^CD-Extra:" nil t) (forward-line 0)
       (setq cd-extra (monk-get-current-line-list))

       (search-forward-regexp "^Album title") (forward-line 0)
       (setq title (monk-get-current-line-list))

       (search-forward-regexp "^T0") (forward-line 0)
       (progn
	 (while (null (looking-at "^Leadout:"))
	   (setq track-list (cons (monk-get-current-line-list) track-list))
	   (forward-line 1))
	 (setq leadout (monk-get-current-line-list)))
       (list tracks
	     discid
	     cd-text
	     cd-extra
	     title
	     (nreverse track-list)
	     leadout)))))

;;
;; These inlines refers to `monk-track-list' (not `monk-disc-info')
;;

(defsubst monk-disc-info-get-track (track)
  "Extract track data indexed by TRACK out of `monk-track-list'."
  (and monk-track-list (nth (1- track) (car monk-track-list))))

(defsubst monk-disc-info-get-length (track)
  "Extract track length in sectors indexed by TRACK out of `monk-track-list'."
  (let ((trackdat (monk-disc-info-get-track track)))
    (when trackdat (nth 3 trackdat))))

(defsubst monk-disc-info-get-length-sec (track)
  "Extract track length in seconds indexed by TRACK out of `monk-track-list'."
  (let ((trackdat (monk-disc-info-get-track track)))
    (when trackdat (monk-cdtime2sec (nth 1 trackdat)))))

(defsubst monk-disc-info-get-offset (track)
  "Extract offset of TRACK in cd media out of `monk-track-list'."
  (let ((trackdat (monk-disc-info-get-track track)))
    (when trackdat (nth 2 trackdat))))

;; Only this one refers also to `monk-disc-info' as failsafe.
(defun monk-get-discid (&optional with-pfx)
  "Get `discid' from current `monk-track-list' or`monk-disc-info'.
Valid value is returned as string."
  (let (raw)
    (and 
     ;; CDDB data is unused.
     (or (setq raw (and monk-track-list (car (last monk-track-list))))
	 (setq raw (and monk-disc-info (nth 1 monk-disc-info))))
     (if with-pfx (nth 2 raw) (substring (nth 2 raw) 2)))))

;;
;; Generic track generator for cdaudio
;; Sat Jan  6 17:09:26 2001
;;

(defun monk-cda-generate-a-track (track)
  "Generate a generic track out of track number TRACK \(must be an integer\).
Calls `monk-disc-info-get-offset' etc., so, monk-track-list. should already
be available."
  (let ((frames (monk-disc-info-get-length track))
	;; This `options' generations are almost the same as in
	;; the old `monk-cda-generate-command':
	(optlis (append	;; (list monk-player-command)
		 (and  ; (string-equal monk-player-command "cdda2wav")
		  (string-match monk-player-command "cdda2wav")
		  (monk-get-device-opt))
		 ;; append user's or nil
		 (if monk-post-filter
		     (delete "-e" monk-command-options) ; failsafe
		   monk-command-options)
		 ;; automatically generate -e
		 (and (null monk-post-filter) '("-e" "-N")))))
    (list (cons 'track-number track)
	  (cons 'track-offset (monk-disc-info-get-offset track))
	  (cons 'title (monk-cddb-get-title track))
	  (cons 'player monk-player-command)
	  ;;(cons 'options (when optlis (mapconcat 'identity optlis " ")))
	  (cons 'options optlis)
	  '(track-opt . "-t%d")
	  '(offset-opt . "-o%d")
	  '(duration-opt . "-d%df")
	  ;; If `monk-post-filter' is non-nil, "-" may have to be generated
	  ;; as 'file property, ie., to support old `c2w2cdp' script.
	  ;; However, there will cause a lot of pain to modify current
	  ;; monk.el to support "-" as a valid file name designating stdout:
	  ;; there are lot of `file-readable-p' and `file-attributes' code.
	  ;; Until a generic `monk-file-check' function is complete,
	  ;; `c2w2cdp' must output to stdout without "-".
	  ;; (cons 'file (and monk-post-filter (list "-")))
	  (cons 'post-filter monk-post-filter)
	  (cons 'frames frames)
	  ;; (cons 'fps monk-cdaudio-fps)
	  (cons 'fps 75.0)		; good enough ?
	  (cons 'time (monk-disc-info-get-length-sec track))
	  (cons 'cent (/ frames 100.00))
	  (cons 'type 'cda))))

;;
;; `monk-cda-insert-a-track' -- One track inserter
;; (monk-cda-insert-a-track 3)
;;  cda:  [04:22.00]    61312    19688 "Aquilarco #3 (Ornithomanteia)"
;;

(defsubst monk-cda-insert-a-track (track)
  "Insert a MonK track for cdaudio TRACK \(integer\)."
  (let (ttl (obj (monk-cda-generate-a-track track)))
    (monk-insert-track obj
		       (format "%s:" (monk-assq-get-value 'type obj))
		       (format "%5s.00" (monk-sec2cdtime
					 (monk-assq-get-value 'time obj)))
		       (monk-assq-get-value 'track-offset obj)
		       (monk-assq-get-value 'frames obj)
		       ;; (setq ttl 
		       (format "\"%s\"" (monk-assq-get-value 'title obj))
		       'newline)
    ;; (message "Monk read Track: %s" ttl)
    obj))

;;
;; New `monk-insert-track-lines' for generic `track' generation
;; Sat Jan  6 18:54:50 2001
;;

(defun monk-insert-track-lines ()
 "Insert cdaudio tracks as MonK `track' object in current buffer."
  (let* ((buffer-read-only)
	 (tracks (length (car monk-track-list)))
	 (tracksstr (car (nth 2 monk-track-list)))
	 (trk 1)
	 (total (nth 1 (nth 2 monk-track-list)))
	 (discid (nth 4 monk-track-list))
	 (atitle (monk-cddb-get-album-title)))
    (widen)
    (if monk-readin-redisplay-every (setq monk-readin-counter 0))
    (while (> tracks 0)
       (monk-cda-insert-a-track trk)
       (cond (monk-readin-redisplay-every
	      (setq monk-readin-counter (1+ monk-readin-counter))
	      (when (= 0 (% monk-readin-counter monk-readin-redisplay-every))
		(sit-for 0))))
      (setq trk (1+ trk)
	    tracks (1- tracks)))
    (insert "\n")
    (and atitle (insert (format "Album:  \"%s\"\n" atitle)))
    (insert (format "DiscID: %s\n" (nth 2 discid)))
    (insert (format "%s\n%s\n" total tracksstr))))

;;;
;;; Cddb --- Get track titles using local or remote CDDB database
;;; See scripts/monk-cddb -- it requires cddbcmd(1) from xmcd(1x).
;;;

;; Wed Nov 22 21:27:26 2000 the routines were imported from monk-jam.el.

(defvar monk-use-cddb-server nil
  "*Defaults nil, meaning remote cddb server is not checked for track titles etc.
It may be safe to set to nil unless cddb server is always available.
For server settings, see `monk-cddb' script, which should be included in
MonK distribution.")

(defconst monk-cddb-buffer " *monk*cddb*"
  "Work buffer to get cddb data.")

(defvar monk-cddb-program "monk-cddb"
  "cddb data fetcher. See the sample script in monk distribution.")

(defvar monk-cddb-data nil "CDDB data list \(flat\).  Eg.,

\(\"DTITLE=Janis Ian / Up 'Til Now\"
 \"DISCID=fa10ef11\"
 \"TTITLE0=Stars\"
  :
 \"TTITLE16=This Night\"
  \)")

(defconst monk-cddb-save-dir "~/monk-cddb")

(defvar monk-cddb-do-save t
  "*Defaults t, meaning save fetched cddb data should be saved in
`monk-cddb-save-dir'.")

(defsubst monk-cddb-generate-fname (discid &optional dir)
  "Generate file name list from string DISCID.
Sometimes cddb data has \"9d0db50c,9b0db30c\" sort of discid, and
the data should be hard linked.
Optional 2nd arg DIR is prepended to each"
  (let* ((discid (dired-replace-in-string " " "" discid)) ; failsafe
	 (files (dired-split "," discid)))
    (mapcar
     (lambda (f)
       (expand-file-name
	(concat (and dir (concat (monk-hack-trailing-/ dir) "/") ) f)))
     files)))

(defsubst monk-cddb-discid-differs-cdda2wav-p ()
  "Compare discid from cdda2wav that and from cddb.
Returns discid of cdda2wav if the two differs, or nil if the two
is identical.  Called from `monk-save-cddb-data'."
  (let ((c2wname (monk-get-discid))
	(cddbname (monk-cddb-get-discid)))
	(cond ((null (and (stringp c2wname) (stringp cddbname))) nil)
	      ((string-equal c2wname cddbname) nil)
	      (t c2wname)))) 

;; Currently, whether the data source of `monk-cddb-program' is local or
;; remote is not checked, so save or not-save should be decided with
;; checking [ -f "$discid" ] in this function.
(defun monk-save-cddb-data ()
  "Save current-buffer content as cddb data in `monk-cddb-save-dir'.
Returns nil if failed, or created file\(s\) as a list.
File may be written in multiple hardlink.
If cdda2wav\(1\) has returned different discid, the name is also
recorded as symbolic link."
  (let (discid files top rest srcf symlink)
    (cond ((and monk-cddb-do-save
		monk-cddb-data
		(setq discid (monk-cddb-get-discid))
		monk-cddb-save-dir
		(and (stringp discid) (stringp monk-cddb-save-dir))
		(file-writable-p monk-cddb-save-dir)
		(setq files (monk-cddb-generate-fname discid
						       monk-cddb-save-dir)
		      rest (cdr files)
		      top (car files)))
	   ;; Failsafe
	   (setq default-directory
		 (monk-put-trailing-/ (expand-file-name monk-cddb-save-dir)))
	   (widen)
	   ;; The 1st file,
	   (and (not (file-exists-p top))
		(write-region (point-min) (point-max) top))
	   ;; then hard link all the names listed in CDDB discid record,
	   (mapcar (lambda (hlink)
		     (and (not (file-exists-p hlink))
			  (add-name-to-file top hlink t)))
		   rest)
	   ;; and, also make cdda2wav discid as symlink.
	   (and (setq symlink (monk-cddb-discid-differs-cdda2wav-p))
		(setq srcf (file-name-nondirectory top))
		(null (string-equal symlink srcf)) ; unique each other ?
		(not (file-exists-p
		      (car (monk-cddb-generate-fname symlink
						      monk-cddb-save-dir))))
		(make-symbolic-link srcf symlink))
	   files))))			; else nil if failed

(defun monk-fetch-cddb-data (&optional qparam)
  "Fetch cddb data for current CD and setup `monk-cddb-data'.
Returns the new `monk-cddb-data'.
Use query parameter string QPARAM as optional arg if it's non-nil."
  (let ((buf (get-buffer-create monk-cddb-buffer)))
    (save-excursion
      (setq default-directory "~/")
      (set-buffer buf) (erase-buffer)
      (sit-for 0) (message "Getting cddb data...")
      (prog1
	  (cond
	   ;; nil is concat()'ed to argument
	   ((= 0 (monk-with-coding-system
		  (call-process shell-file-name nil t nil "-c"
				(format "%s %s %s" monk-cddb-program
					(if monk-use-cddb-server "" "-R")
					qparam))))
	    (message "Getting cddb data...Done")
	    (sit-for 1) (message "")
	    (setq monk-cddb-data (monk-get-cddb-track-data)))
	   (t (message "Can't get CDDB data, sister...")
	      (setq monk-cddb-data nil)))
	;; Post-job:
	;; `cd' to `monk-cddb-save-dir', and save if required
	(and monk-cddb-save-dir
	     ;;(setq default-directory
	     ;;      (expand-file-name monk-cddb-save-dir))
	     (cd (expand-file-name monk-cddb-save-dir))) ; error raised
	(monk-save-cddb-data)))))

(defun monk-get-cddb-track-data-squeeze1 (raw1)
  "Eliminate unwanted record from RAW1.
RAW1 is derived from `monk-cddb-buffer'
Called from `monk-get-cddb-track-data'."
  (delete "" (mapcar (lambda (data)
		       (if (string-match
			    "^DTITLE=\\|^DISCID=\\|^TTITLE[0-9]+=" data)
			   (dired-split "=" data) ""))
		     raw1)))

;; The new squeezer2, `while' version.
(defun monk-get-cddb-track-data-squeeze2 (raw2)
  "Unique's asociated list RAW2.  Returns resulted list in reverse order.
Required, because sometimes a cddb data has this

  TTITLE0=Looooooooooon .......................................
  TTITLE0=g Song Title

sort of multiple line format for one `title' record.

RAW2 is assumed to be a list of data formated as \(keystr valstr\).
If `next' data in RAW having same key, the `next' valstr is
concatenated to `this' valstr to make the list a unique asociated list.
Called from `monk-get-cddb-track-data'."
  (let (this thiskey thisval next nextkey nextval newlis)
    (while raw2
      (setq this (car raw2)
	    thiskey (car this)
	    thisval (car (cdr this))
	    next (car (cdr raw2))
	    nextkey (car next)
	    nextval (car (cdr next)))
      (cond
       ((string-equal thiskey nextkey)
	(setq this (list thiskey (concat thisval nextval))
	      raw2 (cons this (cdr (cdr raw2)))))
       (t (setq newlis (cons this newlis)
		raw2 (cdr raw2)))))
    (nreverse newlis)))

;; Old monk-get-cddb-track-data-squeeze2 (`recursive' version)
;; This barfs if (> (length RAW2) 120 to 130):
(defun monk-get-cddb-track-data-squeeze2-recursive (raw2)
  (let (this thiskey thisval next nextkey nextval)
    (setq this (car raw2)
	  thiskey (car this)
	  thisval (car (cdr this))
	  next (car (cdr raw2))
	  nextkey (car next)
	  nextval (car (cdr next)))
    (cond
     ((null next) raw2)			; (= (length raw2) 1) case.
     ;; (Writing this part was realy pain)
     ;; Yeah, next data has data to be concatenated, so
     ;; Make new `this' and pass it as `car' of re-structured cdr.
     ((string-equal thiskey nextkey)
      (setq this (list thiskey (concat thisval nextval))) ; new this
      (monk-get-cddb-track-data-squeeze2 (cons this (cdr (cdr raw2)))))
     ;; else yield `this'
     (t  (cons this (monk-get-cddb-track-data-squeeze2 (cdr raw2)))))))

(defun monk-get-cddb-track-data ()
  "Scan current buffer and creates data bind-able to `monk-cddb-data'.
Normally executed in `monk-cddb-buffer'.
The data will be unique'd by `monk-get-cddb-track-data-squeeze1' and 
`monk-get-cddb-track-data-squeeze2'."
  (let* ((raw1 (dired-split "\n"
			    (buffer-substring (point-min) (point-max))))
	 (raw2 (monk-get-cddb-track-data-squeeze1 raw1))
	 (newlis (monk-get-cddb-track-data-squeeze2 raw2)))
    ;; Concat'ing again with "=".  Inefficiency.
    ;; Must check if how many functions depends on the flat structure
    ;; of `monk-cddb-data' before I Change `monk-cddb-data' to real alist.
    (mapcar (lambda (data) (mapconcat 'identity data "=")) newlis)))

(defun monk-cddb-build-query-param ()
  "Build query parameter out of current `monk-track-list'."
  (let* ((discidstr (nth 2 (nth 4 monk-track-list)))
	 (tracks-total-str (nth 2 monk-track-list))
	 (tracksstr (car tracks-total-str))
	 (totalstr (car (cdr tracks-total-str)))
	 (tracklis (car monk-track-list))
	 discid tracks offsets total)
    (setq discid (substring discidstr 2)
	  tracks (substring tracksstr (string-match "[0-9]+$" tracksstr))
	  offsets (mapconcat (lambda (tr)
			       (int-to-string ; emacs-21 barfs without this.
				(nth 2 tr))
			       ) tracklis " "))
    (setq total (substring totalstr (string-match ": *\\([0-9:\.]+\\)$"
						  totalstr))
	  total (substring totalstr (match-beginning 1) (match-end 1))
	  total (int-to-string ; emacs-21 barfs without this.
		 (round (monk-cdtime2sec total))))
    (mapconcat 'identity (list discid tracks offsets total) " ")))

(defsubst monk-cddb-string-match (string)
  "Do `string-match' with STRING on `monk-cddb-data'.
All matched data are returned as list.
Do not expect \(match-data\) is always what you want."
 (delq nil (mapcar (lambda (data)
		    (if (string-match string data) data) ) monk-cddb-data)))

(defsubst monk-cddb-get-matched-data (string)
  "Do `monk-cddb-string-match' and return matched record's data part.
Eg., \(monk-cddb-get-matched-data \"DTITLE=\"\)
Do not add \"^\" or \"=\" part.  They are generated automatically."
  (and monk-cddb-data			; this must be checked.
       (stringp string)
       (let ((record (car (monk-cddb-string-match (concat "^" string)))))
	 (cond
	  ((and (stringp record)
		(string-match (concat "^" string "=\\(.*$\\)") record))
	   (substring record (match-beginning 1)))))))

(defun monk-cddb-get-discid ()
  "Get cddb discid."
  (monk-cddb-get-matched-data "DISCID"))

(defun monk-cddb-get-album-title ()
  "Get cddb album title.  Currently only cddb data is used."
  (or 
   (monk-cddb-get-matched-data "DTITLE")
   ;; (nth 2 (nth 3 monk-track-list)) ; Unused.
      ))

;; Skelton derived from old `monk-insert-track-lines-integer' (kept in
;; monk-purged.el), but uses `monk-cddb-get-matched-data'.

(defun monk-cddb-get-title (track)
  "Get cddb track title for TRACK number \(integer\)."
  (let* ((tracklist (car monk-track-list))
	 ;; (title (nth (1- track) (nthcdr 2 monk-cddb-data)))
	 (trk (nth (1- track) tracklist)))
    (or
     (monk-cddb-get-matched-data (format "TTITLE%d" (1- track)))
     (mapconcat 'identity (nthcdr 4 trk) " "))))
