;; `monk-fn-handler.el' --- find-file handler functions for MonK

;; You need MonK (http://www.asahi-net.or.jp/~pi9s-nnb/monk/) to us this.

;; Mon Oct 21 23:57:39 2002 s.n.
;; Copyright 2002 Seiichi Namba <sn@asahi-net.email.ne.jp>

;; monk.el works on emacs-20.7 / emacs-21.2 / emacs-19.34 / emacs-19.28
;; but this module does not support emacs-19.28 yet.

;; DESCRIPTION
;;
;;	This module enables emacs to `find-file' multimedia file such
;;	as having extensions such as .mp3, .mpeg, .wav etc. if your
;;	MonK supports the file format. All the function-call for
;;	`insert-file-contents' are hooked, and the file is `inserted'
;;	as MonK track in a `monk-mode' buffer.  The operation is
;;	enabled on all of the standard emacs `find-file' commands such as:
;;	
;;		C-x C-f (find-file)
;;		C-x 4 f (find-file-other-window)
;;		C-x 5 f (find-file-other-frame)
;;      	f (dired-find-file)                ;; dired-mode
;;		o (find-file-other-window)         ;; dired-mode
;;		C-o (dired-display-file)           ;; dired-mode
;;	
;;	and many other commands. Even invoking emacs from shell
;;
;;		$ emacs ~/mp3/SweetJane.mp3
;;
;;	works as you expect.

;; INSTALL
;;
;;	Load from `monk-load-hook':
;;
;;	(add-hook 'monk-load-hook (lambda () (require 'monk-fn-handler)))
;;
;;	Or just load.  This file loads monk.el if it is not loaded, so
;;	if you just load this file, all your intended `monk-load-hook'
;;	setup must have been all done at loading this file.
;;	Variable `monk-file-handler-auto-start' can be set to nil if
;;	the automatic-play-start facility is annoying.
;;

(provide 'monk-fn-handler)

(require 'monk)
(require 'cl)

(defvar monk-fn-handler-version "0.3")

(defvar monk-file-handler-auto-start t
  "Defaults t, which means `monk-file-handler' starts \
to play the relevant multimedia file.")

;; `file-name-handler-alist' setup.
;; Currently, `monk-track-generator-list' is used as a seed, but this
;; might not be a good method:
;;
;; (1) It might not be good enough since modules such as `monk-mp3.el'
;; etc. is normally loaded afterwards and may contain new handlers.
;;
;; (2) `file-name-handler-alist' may contain much more extensions than
;; that list in `monk-track-generator-list', since finding unsupported
;; file will fail anyway.
;;
;; For debug,
;; (setq file-name-handler-alist-backup (copy-list file-name-handler-alist)) ;; save 
;; (setq file-name-handler-alist (copy-list file-name-handler-alist-backup)) ;; restore

(or (assoc "\\.mp3$" file-name-handler-alist)
    (setq
     file-name-handler-alist
     (append
      file-name-handler-alist
      (delq ()
	    (mapcar
	     (lambda (l) 
	       ;; The functions below can't handle ".m3u" extension yet,
	       ;; since multiple files should be `found' from firs
	       ;; `finding' of initial xxx.m3u.  On the other hand,
	       ;; recursive finding `directory' works as coded below.
	       (if (string-match "m3u" (car l))
		   nil
		 (cons (car l) (function monk-file-handler))))
	     monk-track-generator-list)))))

;; IFS of the function to be hooked. Only 1st one can be useful for us.
;; (insert-file-contents FILENAME &optional VISIT BEG END REPLACE)

(defun monk-file-handler (operator &rest args)
  "File operation handler for inserting multimedia files.
OPERATOR may be expand-file-name, file-attributes, insert-file-contents
etc., and ARGS is to it.  Intended to work under standard
`file-name-handler' system in emacs and should be registered in
`file-name-handler-alist'."
  (cond
   ((eq operator 'insert-file-contents)
    (let ((bs (buffer-size)))
      (when
	  (monk-insert-file (car args)) ; The undertaker blow.
	(setq buffer-read-only t)	; sometimes in vain.
	(when monk-file-handler-auto-start
	  (when (monk-proc-running-p) (monk-stop))
	  (monk-play-this-track) )
	;; Real `insert-file-contents' returns list of absolute file
	;; name and number of bytes inserted, so must fake return vale
	;; as being (). But, `buffer-file-name' fake below is still needed...
	(list (expand-file-name (car args)) (- (buffer-size) bs)))))
   (t
    ;; Else call OPERATOR fun via `monk-file-handler-default'
    (apply (function monk-file-handler-default) operator operator args))))

(defun monk-file-handler-default (function operator &rest args)
  "Call FUNCTION with temporary binding on `inhibit-file-name-handlers' \
using OPERATOR.
Optional argument ARGS are to be passed to FUNCTION."
  (let ((inhibit-file-name-handlers
	 (cons 'monk-file-handler
	       (and (eq inhibit-file-name-operation operator)
		    inhibit-file-name-handlers)))
	(inhibit-file-name-operation operator))
    (apply function args)))

;; Based on `monk-find-file': This does not creates or switch to other
;; `monk-buffer', but just uses current buffer.
;; Currently always called with arg==fakeinteractive==nil.

(defun monk-insert-file (file &optional arg fakeinteractive)
  "Insert multimedia FILE to current buffer, and setup `monk-mode'."
  (interactive "fMonK Insert File: \nP")
  ;; (message "X:%S" (setq ll (list file arg fakeinteractive)))
  (let ((buf (current-buffer))
	(file (expand-file-name file)))
    (save-excursion
      (set-buffer buf)
      ;; (erase-buffer)

      (monk-with-coding-system
	(let (handler)
	  (cond
	   ((file-directory-p file)
	    (and (or (interactive-p) fakeinteractive) (cd file))
	    (monk-find-directory file arg) ; Recurrs to this...
	    (setq buffer-file-name 
		  (generate-new-buffer-name monk-buffer-name) ; The fake
		  buffer-read-only t)
	    )
	   ;; Else take just one file.
	   (t
	    (cond
	     ;; file extension is tested here
	     ((setq handler (monk-get-track-generator file))
	      ;; Now use current-buffer
	      (cond
	       ((file-readable-p file)
		(cond
		 ((= 0 (nth 7 (file-attributes file)))
		  (error "Can't 0-sized file %s" file))
		 ;; Without `monk-with-monk-window' here.
		 (t 
		  (let (buffer-read-only)
		    ;; `monk-with-no-face' disables
		    ;; `monk-file-handler-auto-start' operation.
		    ;; So
		    (progn ;; monk-with-no-face
		     (forward-line 0)
		     ;; Call the track generation handler
		     (cond
		      ((funcall handler file)
		       (monk-readin-redisplay nil)
		       (set-buffer-modified-p nil)
		       (message "MonK read file \"%s\"" file)
		       ;; Must update this list:
		       (setq monk-buffer-list
			     (append monk-buffer-list (list buf)))
		       (setq buffer-file-name file ; The fake.
			     buffer-read-only t) ; This sets in vain.
		       (monk-mode)
		       t)
		      (t (error "Probe failed for file %s" file) nil)
		      )))
		  )))
	       (t (error "Can't read file %s" file) nil)))
	     (t (error "Can't find handler for %s" file)))))))
      )))

;; Maybe not so useful

(defun monk-find-file2 (file &optional arg)
  "Variation of `monk-find-file'. Do not seek for MonK buffer."
  (interactive "fMonK find File: \nP")
  (cond ((file-readable-p file)
	 (monk-open-new)
	 (monk-insert-file file arg))
	(t (error "Can't read file %s" file) nil)))
