;; monk-mp3.el --- mp3 track generator module for MonK

;; (c) 2000,2001,2002,2003 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; Started: Mon Dec  4 01:57:31 2000 s.n.

;;
;; What we do in this file is:
;;
;; (1) To build this sort of alist as a `track' object out of `file':
;;
;; ((player . "mpg123")
;;  (track-opt)
;;  (offset-opt . "-k%d")
;;  (duration-opt . "-n%d")
;;  (post-filter)
;;  (file . "~/mp3/Classical/Beethoven - Piano Sonata No.14-1-.mp3")
;;  (frames . 13968)
;;  (time . 364.87)
;;  (cent . 139.68)
;;  (fps . 38)
;;  (type . mp3)
;;  (title . "Moonlight - 1.Adagio Sostenuto")
;;  (artist . "Alfred Brendel")
;;  (album . "")
;;  (year . "")
;;  (comment . "")
;;  (genre . "Classical"))
;;
;; (2) To insert the `track' in this sort form:
;;
;; mp3:  [06:04.00]        0    13968 `Moonlight - 1.Adagio Sostenuto - Alfred Brendel'
;;

;;
;; There's a lot of pain to draft this file:
;;
;; (1) Without stdout-redirection-to-/dev/null, str search fails.
;;     Done by ">/dev/null" and `shell-file-name' in the old
;;     `monk-mp3-get-info' in old versions, but this is now done by
;;     mpg123's option "-t".
;; (2) Can't redirect stdout and get stderr without `shell-file-name'
;;     in emacs-19, which we should still support for a little while.
;;     This is remedied by the option "-t" as well.

;; (3) If `post-filter' is non-nil, shell is inevitable to create
;;     a pipe, and had to use `dired-shell-stuff-it' for mp3 filename.
;;     It works most of cases, but I don't know what happens if the file
;;    is from Macintosh/Windows (via napster) and having SJIS-encoded.
;; Async process is also tried, but it causes another pain, you know.
;;

(require 'monk)
(require 'dired-aux)
(provide 'monk-mp3)

(defvar monk-mp3-player-command "mpg123" "Should be \"mpg123\".")
(defvar monk-mp3-player-command-options nil "*For `monk-mp3-player-command'")
(defvar monk-mp3-probe-command "mpg123"  "Should be \"mpg123\".")
(defvar monk-mp3-player-track-opt nil)
(defvar monk-mp3-player-offset-opt "-k%d" "mpg123\(1\) proper.")
(defvar monk-mp3-player-duration-opt "-n%d" "mpg123\(1\) proper.")
(defvar monk-mp3-player-post-filter nil)

;; New probe option for the new `monk-mp3-get-info':
;; (defvar monk-mp3-probe-options '("-s" "-v" "-k0" "-n10")
;;   "Used to get info from mp3 file.") ;; old one for `shell-file-name' version.

(defvar monk-mp3-probe-options '("-vt"  "-k0" "-n20")
 "Used to get info from mp3 file.
Automatically re-calibrated by `monk-mp3-calibrate-probe-options'.
Do not modify manually.")
;; (setq monk-mp3-probe-options '("-vt" "-k0" "-n100"))
;; (setq monk-mp3-probe-options '("-vt" "-k0" "-n50"))
;; (setq monk-mp3-probe-options '("-vt" "-k0" "-n5"))
;; (setq monk-mp3-probe-options '("-vt" "-k0" "-n0"))

(defvar monk-mp3-extra-frame
  0 "Number of extra frame to be appended to `monk-mp3-probe-options' \
as an argument of \"-n\" when `monk-mp3-build-track-obj' is failed
by `monk-mp3-calibrate-probe-options'.
See also `monk-mp3-read-retry'.")

(defvar monk-mp3-extra-frame-delta
  50 "*Number of extra frame to be added to `monk-mp3-extra-frame' \
by `monk-mp3-calibrate-probe-options', when `monk-mp3-build-track-obj' is failed.
See also `monk-mp3-read-retry'.")

(defvar monk-mp3-extra-frame-max
  1000 "*Maximal value of `monk-mp3-extra-frame'.
The limitation is done by `monk-mp3-calibrate-probe-options'.")

(defvar monk-mp3-readin-wait 0.05
  "A delay in second inserted after reading .mp3 file.
Defaults 0.05 considering mpg123, but can be set to 0 for mpg321 as
`monk-mp3-probe-command'.  0 might make track generation twice faster.")

(defun monk-mp3-get-info (file)
  "Get info including track timing, frames etc. from FILE with mpg123\(1\).
Current version does not use shell, `dired-shell-stuff-it', and \">/dev/null\"
anymore.  See `monk-mp3-probe-command', and `monk-mp3-probe-options'."
  (let ((buf (get-buffer-create monk-info-buffer)) tem)
    (save-excursion
      (set-buffer buf) (erase-buffer)
      (setq default-directory
	    (monk-put-trailing-/ (file-name-directory file)))
      (cond
       ( ;; (= 0 ;; Sometimes "Hangup" is returned in E19.28 instead of 0
	(setq tem ;; hack `tem' if debug required.
	    (apply 'call-process monk-mp3-probe-command nil t nil
			    (append monk-mp3-probe-options
				    (list
				     file ; expanded already
				     ))))
	;; Delay. Recursive directory read sometimes failed.
	;; mpg123 is hard to deal with.  Must be put here (or where ?)
	;; (sleep-for (/ monk-wait-proc-to-die 2)) ; may be too short...
	(sleep-for monk-mp3-readin-wait)
	(cons (cons (quote file) file)
	      (monk-mp3-parse-info buf)))))))

;; 
;; `monk-mp3-parse-info' -- should be re-written-maybe-later
;; (1) The regex is too hairy.
;; (2) Maybe we should split the lines into list, then hack it,
;;     like the new functions in `monk-smpeg.el' / `monk-midi.el'
;; 

(defun monk-mp3-parse-info (buf)
  "Parse BUF, to get info from mpg123 output."
  (let ((RegEx
	 ;; Bracket'ed frame# may preceeded by SPC's
	 "^Frame# +\\([0-9]+\\) +\\[ *\\([0-9]+\\)\\ *], +Time: +\\([0-9]+:[0-9]+\\.[0-9]+\\)+ \\[ *\\([0-9]+:[0-9]+\\.[0-9]+\\) *\\]")
	frames id3data specdata)
    (save-excursion
      (set-buffer buf)
      (goto-char (point-min))
      (save-excursion
	(while (search-forward "\r" nil t)
	  (replace-match "\n")))

      (goto-char (point-max))		;; to use the last occurrence
      (cond
       ( ;; (search-forward-regexp RegEx nil t) ; use the first occurrence
	(search-backward-regexp RegEx nil t) ; use the last occurrence

	(setq id3data (monk-mp3-get-id3)
	      specdata (monk-mp3-get-spec))
	(append
	 (list 
	  (cons
	   'frames
	   (setq frames
		 (+ ;; - ;; BIG BUG FIXED
		  (string-to-int
		   (buffer-substring (match-beginning 2) (match-end 2)))
		  (string-to-int
		   (buffer-substring (match-beginning 1) (match-end 1))))))
	  (cons
	   'time
	   (+ ;; - ;; BIG BUG FIXED
	    ;;(save-match-data)
	    (monk-cdtime2sec
	     (buffer-substring (match-beginning 4) (match-end 4)))
	    ;;(save-match-data)
	    (monk-cdtime2sec
	     (buffer-substring (match-beginning 3) (match-end 3)))))
	  (cons 'cent (/ frames 100.0))
	  (cons 'fps 38)
	  ;; (cons 'file file) ;; will be generated by `monk-mp3-get-info'
	  (cons 'type 'mp3))
	 id3data
	 specdata))))))

(defun monk-mp3-get-spec ()
  "Get \"Directory:\" to \"Audio:\" as spec000, spec001..."
  (save-excursion
    (goto-char (point-min))
    (let (p0 p1 splines (count 0) lis brate)
      (save-match-data
	;; (search-forward-regexp "^Directory" nil t) ;; redundant
	(search-forward-regexp "^MPEG *[0-9]" nil t)
	(forward-line 0)
	(setq p0 (point))
	(search-forward-regexp "^$" nil t)
	(setq p1 (1- (point)))
	(setq splines (dired-split "\n" (buffer-substring p0 p1)))
	(mapc
	 (lambda (d)
	   (setq lis (cons
		      (cons
		       (make-symbol
			(format "spec%03d" (setq count (1+ count)))) d) lis))
	   (when (string-match "^Bitrate: *\\([0-9]+\\) *Kbits/s" d)
	     (setq brate (cons
			  'bitrate
			  (string-to-number
			   (substring d (match-beginning 1) (match-end 1))))))
	   )
	 splines)
	(cons brate (nreverse lis))))))

(defconst monk-empty-id3-data		; id3tool(1) option
  '((title)				; -t
    (artist)				; -r
    (album)				; -a
    (year)				; -y
    (comment)				; -n
    (genre))				; -G (string)
  "Empty Id3 data. Used by `monk-mp3-get-id3'.")

(defun monk-mp3-get-id3 ()
  "Gets id3 data from current buffer containing mpg123 output.
With saving `match-data' and `excurstion'."
  (save-excursion
    (goto-char (point-min))
    (let ((matchofs (list 1 2 3 4 5 6))
	  (prop   (list 'title 'artist 'album 'year 'comment 'genre)))
      (save-match-data
	(cond ((search-forward-regexp
		"^Title  : \\(.*\\)  +Artist: \\(.*\\)
Album +: \\(.*\\)  +Year  : \\(.*\\)
Comment: \\(.*\\)  +Genre : \\(.*\\)" nil t)
	       (mapcar
		(lambda (n)
		  (prog1
		      (cons
		       (car prop)
		       (let ((tem (buffer-substring
				   (match-beginning n) (match-end n))))
			 (save-match-data
			   (if (string-match " +$" tem)
			       (substring tem 0 (match-beginning 0)) tem))))
		    (setq prop (cdr prop))))
		matchofs))
	      (t (copy-list monk-empty-id3-data))))))) ; enable user-edit

(defun monk-mp3-build-track-obj (file)
  "Get info from mp3 file FILE, and build track object for the FILE.
`monk-mp3-get-info' is called."
  (let ((info (monk-mp3-get-info (expand-file-name file))))
    (when info
      (append (list
	       (cons 'player monk-mp3-player-command)
	       (cons 'track-opt monk-mp3-player-track-opt)
	       (cons 'offset-opt monk-mp3-player-offset-opt)
	       (cons 'duration-opt monk-mp3-player-duration-opt)
	       (cons 'options monk-mp3-player-command-options)
	       (cons 'post-filter monk-mp3-player-post-filter))
	      info))))

;; Now in monk.el.
;; (defsubst monk-assq-get-value (prop proplis)
;;   (cdr (assq prop proplis)))

(fset 'monk-mp3-insert-mp3-file 'monk-mp3-insert-track-obj)

(defvar monk-mp3-show-title
  t "*If non-nil, Use id3 title info is displayed instead of mp3 filename.
`monk-mp3-insert-track-obj' refers this.")

(defvar monk-mp3-read-retry 10
  "*Retry count allowed to build one mp3 track by `monk-mp3-insert-track-obj'.")

(defsubst monk-mp3-calibrate-probe-options ()
  "Auto-re-calibrate `monk-mp3-probe-options'.
Increase `monk-mp3-extra-frame' by `monk-mp3-extra-frame-delta' and generate
new a \"-n\" option string, and replace the corresponding option string
in `monk-mp3-probe-options'.  The maximum of `monk-mp3-extra-frame' is limited
to `monk-mp3-extra-frame-max'.  Called from `monk-mp3-insert-track-obj',
and very mpg123-proper."
  (setq monk-mp3-extra-frame (+ monk-mp3-extra-frame monk-mp3-extra-frame-delta)
	monk-mp3-extra-frame (cond
			      ((> monk-mp3-extra-frame monk-mp3-extra-frame-max)
			       monk-mp3-extra-frame-max)
			      (t monk-mp3-extra-frame))
	;; Old code in 7.18 just appended the new "-nXXX", but now replace.
	monk-mp3-probe-options (mapcar
				(lambda (o)
				  (cond ((string-equal "-n" (substring o 0 2))
					 (format "-n%d" monk-mp3-extra-frame))
					(t o))) monk-mp3-probe-options)))

(defun monk-mp3-insert-track-obj (file)
  "Insert track line at point.
Tries to build / insert track `monk-mp3-read-retry' times."
  ;; (interactive "fmp3 File:")
  (let (obj fname frames time title artist brate)
    (cond
     ((let ((count 1))
	(catch 'success
	  (while (<= count monk-mp3-read-retry)
	    (when (< 1 count)
	      (message
	       "MonK retrying \(count: %02d\) to read file: %s" count file))
	    ;; `frames' and `time' is mandatory, but 
	    ;; `monk-mp3-build-track-obj' often fails to set these.
	    (setq obj (monk-mp3-build-track-obj file)
		  frames (monk-assq-get-value 'frames obj)
		  time (monk-assq-get-value 'time obj)
		  count (1+ count))
	    (cond
	     ((and obj frames time) (throw 'success obj))
	     (t (monk-mp3-calibrate-probe-options)))) nil))
      (setq ;; pt (point)
	    fname (file-name-nondirectory (monk-assq-get-value 'file obj))
	    title (monk-assq-get-value 'title obj)
	    artist (monk-assq-get-value 'artist obj)
	    brate (monk-assq-get-value 'bitrate obj))
      (monk-insert-track
       obj
       "mp3:"
       (format "%5s.00" (monk-sec2cdtime time))
       (if brate
	   (format "%dk/s" brate)
	 "unknown")
       frames
       (or (cond ((and monk-mp3-show-title title
		       (null (string-equal title "")))
		  (format "`%s%s'" title
			  (cond (artist (format " - %s" artist))
				(t "")))))
	   (format "\"%s\"" fname)) 'newline)
      obj)
     (t (message "Track build faild for file: %s" file)))))

(defun monk-mp3-mouse-toggle-show-title ()
  "Toggle track names between mp3 title and file name.
See also `monk-mp3-toggle-show-title'."
  (interactive)
  (monk-mp3-toggle-show-title nil))

(defun monk-mp3-toggle-show-title (arg)
  "If ARG is nil, toggles `monk-mp3-show-title'.
If numeric ARG is positive, `monk-mp3-show-title' is set to t, else nil."
  (interactive "P")
  (let ((ov monk-mp3-show-title))
    (cond (arg (cond ((> (prefix-numeric-value arg) 0)
		      (setq monk-mp3-show-title t))
		     (t (setq monk-mp3-show-title nil))))
	  (t (setq monk-mp3-show-title (null monk-mp3-show-title))))
    (or (eq monk-mp3-show-title ov) (monk-rebuild-all-tracks t))))

;;
;; Find .m3u file -- moved from monk.el
;;

(defsubst monk-m3u-2-flist (file)
  "Returns \(coding-system file-name-list\) out of m3u FILE."
  (if (monk-buffer-live-p monk-junk-buffer)
      (kill-buffer monk-junk-buffer))
  (let ((buf (get-buffer-create monk-junk-buffer)))
    ;; (save-excursion (set-buffer buf)
    (monk-with-current-buffer
     buf (erase-buffer)
     (insert-file-contents file)
     (cons
      (and (boundp 'buffer-file-coding-system) buffer-file-coding-system)
      (delete
       "" (dired-split
	   "\n" (buffer-substring (point-min) (point-max))))))))

(defun monk-m3u-find-file (file)
  "Opens .m3u file, and read-in the files contained in it."
  ;; (interactive "f.m3u file: \nP")
  (setq file (expand-file-name file))
  ;; (monk-with-coding-system ; not needed. `monk-find-file' does
  (cond
   ((file-readable-p file)
    (let* ((result (monk-m3u-2-flist file))
	   (coding (car result))
	   (files (cdr result)))
      (cond (files			; result should not be used (?)
	     (let ((monk-coding-system (if coding coding monk-coding-system)))
	       (push-mark (point))
	       (mapcar 'monk-find-file files))))))))

;;
;; monk-id3.el --- write id3 info
;;
;; Requires id3tool(1) to write id3 data.
;;
;; Merged:  Sun Jan 28 22:23:12 2001
;; Started: Mon Jan 22 03:51:08 2001, s.n.
;;

;; (require 'monk)
;; (require 'monk-mp3)
;; (require 'monk-prop-edit)
;; (provide 'monk-id3)

;;
;; `monk-write-id3-info' -- For `prop-edit', and `monk-mode' mode:
;;

(defun monk-write-id3-info ()
  "Write the id3 info for current track.
Can be called from `monk-mode' and `monk-prop-edit-mode'."
  (interactive)
  (let* ((in-monk-buffer
	  (or (eq major-mode 'monk-mode) (eq major-mode 'monk-edit-mode)))
	 (track (cond (in-monk-buffer (monk-get-track 'raw))
		      ((eq major-mode 'monk-prop-edit-mode)
		       (save-excursion (goto-char (point-min))
				       (read (current-buffer))))))
	 buf args ret)
    (or (eq (monk-assq-get-value 'type track) 'mp3)
	(error "This track is not `mp3' data"))
    (setq args (monk-generate-id3tool-command track))
    (cond
     (args
      (or (y-or-n-p "The mp3 file will be modified. are you sure ? ")
	  (error "Aborted"))
      (setq buf (get-buffer-create "*monk*error"))
      (monk-with-current-buffer buf (erase-buffer))
      (monk-with-coding-system
       (setq ret (apply 'call-process "id3tool" nil buf nil args))))
     (t (error "MonK couldn't generate command for track: %s" track)))
    (cond
     ((= ret 0)
      (cond
       ((eq major-mode 'monk-prop-edit-mode)
	(monk-prop-edit-finish)
	(when monk-rebuild-track-auto (monk-rebuild-track)))
       (in-monk-buffer (when monk-rebuild-track-auto (monk-rebuild-track))))
      (kill-buffer buf))
     (t
      (display-buffer buf)
      (error
       "id3tool failed. Use %s for `genre' keyword"
       (if (setq tem (where-is-internal 'monk-complete-prop))
	   (format "%s (monk-complete-prop)" (key-description (car tem)))
	 "command `monk-complete-prop'"))))
    ret))

(defun monk-generate-id3tool-command (&optional track)
  "Generate a command line args as list, to set id3 info by id3tool\(1\).
Non-nil optional arg TRACK is used, or current track at point in *monk*."
  (let ((track (if track track (monk-get-track)))
	file k v lis)
    (mapcar (lambda (p)
	      (setq k (car p) v (or (cdr p) ""))
	      (cond ((eq 'title k)   (setq lis (append lis (list "-t" v))))
		    ((eq 'artist k)  (setq lis (append lis (list "-r" v))))
		    ((eq 'album k)   (setq lis (append lis (list "-a" v))))
		    ((eq 'year k)    (setq lis (append lis (list "-y" v))))
		    ((eq 'comment k) (setq lis (append lis (list "-n" v))))
		    ;; Must check legal Genre word.
		    ;; Or write fails !!!
		    ((eq 'genre k)   (setq lis (append lis (list "-G" v))))
		    ((eq 'file k)    (setq file v))))
	    track)
    ;; No `dired-shell-stuff-it' here, no shell used for id3tool(1).
    (when (and lis file) (append lis (list file)))))

(defun monk-show-id3-genre-table (arg)
  "Show `monk-id3-genre-table'.
If with non-nil arg, the table is sorted by genre string."
  (interactive "P")
  (let ((buf (get-buffer-create "*monk*id3*genre*table*"))
	(table 
	 (cond (arg (sort (copy-list monk-id3-genre-table)
			  (lambda (a b) (string< (car a) (car b)))))
	       (t monk-id3-genre-table))))
    (monk-with-current-buffer
     buf
     (erase-buffer)
     (insert (format "%30s %-4s   %3s\n"
		     "Genre" "hex" "dec"))
     (insert (format "%30s %4s   %3s\n"
		     "------------" "----" "---"))
     (mapcar (lambda (en)
	       (insert
		(format
		 "%30s 0x%02x   %3d\n"
		 (car en)
		 (cdr en) (cdr en))))
	     table)
     (setq goal-column 30))
    (monk-show-buffer buf)))

;; (assoc "Other" monk-id3-genre-table) => ("Other" 12)

(defvar monk-id3-genre-table
 '(("Blues"		.	?\x00)
   ("Classic Rock"	.	?\x01)
   ("Country"		.	?\x02)
   ("Dance"		.	?\x03)
   ("Disco"		.	?\x04)
   ("Funk"		.	?\x05)
   ("Grunge"		.	?\x06)
   ("Hip-Hop"		.	?\x07)
   ("Jazz"		.	?\x08)
   ("Metal"		.	?\x09)
   ("New Age"		.	?\x0A)
   ("Oldies"		.	?\x0B)
   ("Other"		.	?\x0C)
   ("Pop"		.	?\x0D)
   ("R&B"		.	?\x0E)
   ("Rap"		.	?\x0F)
   ("Reggae"		.	?\x10)
   ("Rock"		.	?\x11)
   ("Techno"		.	?\x12)
   ("Industrial"	.	?\x13)
   ("Alternative"	.	?\x14)
   ("Ska"		.	?\x15)
   ("Death Metal"	.	?\x16)
   ("Pranks"		.	?\x17)
   ("Soundtrack"	.	?\x18)
   ("Euro-Techno"	.	?\x19)
   ("Ambient"		.	?\x1A)
   ("Trip-Hop"		.	?\x1B)
   ("Vocal"		.	?\x1C)
   ("Jazz+Funk"		.	?\x1D)
   ("Fusion"		.	?\x1E)
   ("Trance"		.	?\x1F)
   ("Classical"		.	?\x20)
   ("Instrumental"	.	?\x21)
   ("Acid"		.	?\x22)
   ("House"		.	?\x23)
   ("Game"		.	?\x24)
   ("Sound Clip"	.	?\x25)
   ("Gospel"		.	?\x26)
   ("Noise"		.	?\x27)
   ("Alt. Rock"		.	?\x28)
   ("Bass"		.	?\x29)
   ("Soul"		.	?\x2A)
   ("Punk"		.	?\x2B)
   ("Space"		.	?\x2C)
   ("Meditative"	.	?\x2D)
   ("Instrumental Pop"	.	?\x2E)
   ("Instrumental Rock"	.	?\x2F)
   ("Ethnic"		.	?\x30)
   ("Gothic"		.	?\x31)
   ("Darkwave"		.	?\x32)
   ("Techno-Industrial"	.	?\x33)
   ("Electronic"	.	?\x34)
   ("Pop-Folk"		.	?\x35)
   ("Eurodance"		.	?\x36)
   ("Dream"		.	?\x37)
   ("Southern Rock"	.	?\x38)
   ("Comedy"		.	?\x39)
   ("Cult"		.	?\x3A)
   ("Gangsta"		.	?\x3B)
   ("Top 40"		.	?\x3C)
   ("Christian Rap"	.	?\x3D)
   ("Pop/Funk"		.	?\x3E)
   ("Jungle"		.	?\x3F)
   ("Native US"		.	?\x40)
   ("Cabaret"		.	?\x41)
   ("New Wave"		.	?\x42)
   ("Psychadelic"	.	?\x43)
   ("Rave"		.	?\x44)
   ("Showtunes"		.	?\x45)
   ("Trailer"		.	?\x46)
   ("Lo-Fi"		.	?\x47)
   ("Tribal"		.	?\x48)
   ("Acid Punk"		.	?\x49)
   ("Acid Jazz"		.	?\x4A)
   ("Polka"		.	?\x4B)
   ("Retro"		.	?\x4C)
   ("Musical"		.	?\x4D)
   ("Rock & Roll"	.	?\x4E)
   ("Hard Rock"		.	?\x4F)
   ("Folk"		.	?\x50)
   ("Folk-Rock"		.	?\x51)
   ("National Folk"	.	?\x52)
   ("Swing"		.	?\x53)
   ("Fast Fusion"	.	?\x54)
   ("Bebob"		.	?\x55)
   ("Latin"		.	?\x56)
   ("Revival"		.	?\x57)
   ("Celtic"		.	?\x58)
   ("Bluegrass"		.	?\x59)
   ("Avantgarde"	.	?\x5A)
   ("Gothic Rock"	.	?\x5B)
   ("Progressive Rock"	.	?\x5C)
   ("Psychedelic Rock"	.	?\x5D)
   ("Symphonic Rock"	.	?\x5E)
   ("Slow Rock"		.	?\x5F)
   ("Big Band"		.	?\x60)
   ("Chorus"		.	?\x61)
   ("Easy Listening"	.	?\x62)
   ("Acoustic"		.	?\x63)
   ("Humour"		.	?\x64)
   ("Speech"		.	?\x65)
   ("Chanson"		.	?\x66)
   ("Opera"		.	?\x67)
   ("Chamber Music"	.	?\x68)
   ("Sonata"		.	?\x69)
   ("Symphony"		.	?\x6A)
   ("Booty Bass"	.	?\x6B)
   ("Primus"		.	?\x6C)
   ("Porn Groove"	.	?\x6D)
   ("Satire"		.	?\x6E)
   ("Slow Jam"		.	?\x6F)
   ("Club"		.	?\x70)
   ("Tango"		.	?\x71)
   ("Samba"		.	?\x72)
   ("Folklore"		.	?\x73)
   ("Ballad"		.	?\x74)
   ("Power Ballad"	.	?\x75)
   ("Rhytmic Soul"	.	?\x76)
   ("Freestyle"		.	?\x77)
   ("Duet"		.	?\x78)
   ("Punk Rock"		.	?\x79)
   ("Drum Solo"		.	?\x7A)
   ("Acapella"		.	?\x7B)
   ("Euro-House"	.	?\x7C)
   ("Dance Hall"	.	?\x7D)
   ("Goa"		.	?\x7E)
   ("Drum & Bass"	.	?\x7F)
   ("Club-House"	.	?\x80)
   ("Hardcore"		.	?\x81)
   ("Terror"		.	?\x82)
   ("Indie"		.	?\x83)
   ("BritPop"		.	?\x84)
   ("Negerpunk"		.	?\x85)
   ("Polsk Punk"	.	?\x86)
   ("Beat"		.	?\x87)
   ("Christian Gangsta Rap" .	?\x88)
   ("Heavy Metal"	.	?\x89)
   ("Black Metal"	.	?\x8A)
   ("Crossover"		.	?\x8B)
   ("Contemporary Christian" .	?\x8C)
   ("Christian Rock"	.	?\x8D)
   ("Merengue"		.	?\x8E)
   ("Salsa"		.	?\x8F)
   ("Trash Metal"	.	?\x90)
   ) "Id3 genre keywords and corresponding code")

;; Local Variables:
;; truncate-lines: t
;; end:
