;;; monk-ogg.el --- OggVorbis track generator module for MonK
;;; monk-ogg.el --- New probe for .ogg, no sh script used

;; (c) 2001,2002,2003 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; Started: Thu May 10 00:53:14 2001, s.n.
;; New 0.6: Tue Nov 12 21:02:40 2002, s.n.
;;     0.7: Fri Nov 15 00:56:18 2002, s.n.

;; BUGS: 
;;    (1) vorbiscomment is not supported (yet).

;;
;; The track to be built by this module looks like:
;;
;; ((file . "~/mp3/Diana_Darby-Ragdoll.ogg.mp3")
;;  (player . "ogg123")
;;  (track-opt)
;;  (offset-opt . "-k%s")
;;  (duration-opt)
;;  (options)
;;  (post-filter)
;;  (frames . 226.15)
;;  (time . 226.15)
;;  (cent . 2.2615)
;;  (fps . 1)
;;  (type . ogg))
;;
;; As you can see, `time' property should be used as frames, and to compute
;; `cent' (frames / 100).  This is mostly because ogg123(1) requires that the
;; argument for "-k" should be in second.  Accordingly `fps' is 1.
;;

;; Currently assumed that ogg123(1) should generate output:
;; Ogg123 from vorbis-tools 1.0beta4, Ogg123 from vorbis-tools 1.0rc1:
;; "Time: 00:00.01 [03:07.01] of 03:07.03, Bitrate: 301.4   "
;; ogg123 from vorbis-tools 1.0rc3:
;; "Time: 00:00.09 [03:06.93] of 03:07.03  (112.5 kbps)  Output Buffer   3.1% "

(require 'monk)
(require 'dired)
(require 'dired-aux)
(provide 'monk-ogg)

(defconst monk-ogg-version "0.7")

(defvar monk-ogg-info-str nil
  "Internal use only.  Do not modify.
If `monk-ogg-probe' succeeded, track timing bit-rate string stored.
Set to -1 if the probe failed.")

;; monk-ogg.el (0.5) default settings:
;; (setq monk-ogg-probe-command "/usr/local/bin/ogg123")
;; (setq monk-ogg-probe-command "/usr/bin/ogg123")
;; (setq monk-ogg-probe-command "/slack80/usr/bin/ogg123")
(defvar monk-ogg-probe-command "ogg123")
(defvar monk-ogg-probe-options '("-vd" "null"))

(defvar monk-ogg-player "ogg123" "Should be \"ogg123\".")
(defvar monk-ogg-player-options nil "For `monk-ogg-player'")

(defvar monk-ogg-player-offset-opt "-k%s")
(defvar monk-ogg-player-duration-opt nil)
(defvar monk-ogg-player-post-filter nil)

(defvar monk-ogg-readin-wait 0
  "A delay in second inserted after reading .ogg file.")

(defun monk-ogg-probe-filter (proc str)
  "A process filter for .ogg file probe. Set by `monk-ogg-probe'."
  (monk-with-current-buffer monk-info-buffer (insert str)) ; the buf exists here.
  ;; (monk-with-current-buffer (get-buffer monk-info-buffer) (insert str))
  ;;(message "monk-ogg-probe-filter: %S" str)
  (let ((re0 "Time: +[0-9]")
	(re1 "\\(Time:.*\\)[\n\r]")	; \r appears in emacs-21.2
	;; ogg123 does not raise correct unix return code. What a crap.
	(ree0 "not an Ogg")		; Error by 1.0beta4, 1.0rc1
	(ree1 "may be corrupted"))	; Error by 1.0rc3
    (save-match-data
      (cond
       ((string-match re0 str)
	(when (string-match re1 str)
	  (interrupt-process proc)
	  (setq monk-ogg-info-str
		(substring str (match-beginning 1) (match-end 1)) )))
       ((or (string-match ree0 str) (string-match ree1 str))
	(interrupt-process proc)	; maybe exited already.
	(message "monk-ogg-probe-filter: %S" str)
	;; Must be non-nil to stop `monk-ogg-probe'
	(setq monk-ogg-info-str -1))))))

(defun monk-ogg-junk-sentinel (proc stat)
  "A sentinel for .ogg file probe. Nop for now. 
Used to suppress the silly message `Process ogg-junk finished'
in `monk-info-buffer'."
  ;;(message "sentinel:%S:%S:" proc stat)
  ())

(defvar monk-ogg-probe-idle 0.3
  "Internal idling time required for probing .ogg file in seconds.
You are not expected to understand this. Neither do I.")

;;
;; Asynchronous probe, `sit-for' is kinda tricky
;;

(defsubst monk-ogg-probe (file)
  "Probe FILE as it were .ogg file"
  (let ((buf (get-buffer-create monk-info-buffer))
	(size (nth 7 (file-attributes file)))
	(monk-ogg-probe-idle monk-ogg-probe-idle) ; use temp binding
	(continue t) proc (i 0))
    (setq monk-ogg-info-str nil)
    (monk-with-current-buffer buf (erase-buffer))
    (when (> size 8000000)		; Adjusted by `size'
      (setq monk-ogg-probe-idle (* 2 monk-ogg-probe-idle)))
    (setq proc
	  (apply 'start-process "ogg-junk" buf monk-ogg-probe-command
		 (append monk-ogg-probe-options (list file))))
    (set-process-filter proc 'monk-ogg-probe-filter)
    (set-process-sentinel proc 'monk-ogg-junk-sentinel)
    (sit-for monk-ogg-probe-idle)	; Why does this work ?
    (while continue
;;;      (message "%04d:%S:%S" (setq i (1+ i))
;;;	       monk-ogg-info-str (process-status proc))
      (when monk-ogg-info-str
	(setq continue nil)))
    monk-ogg-info-str))

;;
;; Parser supporting ogg123-1.0beta4, 1.0rc1, and 1.0rc3
;;
;;
;; 1.0beta4, 1.0rc1:
;;      "Time: 00:00.01 [03:07.01] of 03:07.03, Bitrate: 301.4   "
;; 1.0rc3:
;;      "Time: 00:00.02 [03:07.00] of 03:07.03  (182.4 kbps)  Output Buffer   0.0% "
;;

(defsubst monk-ogg-parse-info-str (str)
  "Parse buf and generate ogg track data. Called by `monk-ogg-get-info'."
  (let ((regex0				; Parser for 1.0beta4, 1.0rc1
	 "^Time: +\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\) +\\[\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\)\\] +of +\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\), +Bitrate: +\\([0-9\.]+\\)")
	(regex1				; Parser for 1.0rc3
	 "^Time: +\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\) +\\[\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\)\\] +of +\\([0-9][0-9]:[0-9][0-9]\\.[0-9][0-9]\\) +(\\([ 0-9\.]+\\) +kbps)")
	time)
    (save-match-data
      (when (or (string-match regex0 str) (string-match regex1 str))
	(list
	 (cons 'time
	       (setq time
		     (monk-cdtime2sec
		      (substring
		       str (match-beginning 3) (match-end 3)))))
	 (cons 'frames time)
	 (cons 'cent (/ time 100.0))
	 (cons 'fps 1)
	 (cons 'type 'ogg)
	 (cons 'bitrate
	       (string-to-number
		(substring str (match-beginning 4) (match-end 4)))))))))

;;
;; Brand new in 0.6, uses the new function `monk-ogg-probe'.
;;

(defun monk-ogg-get-info (file)
  "Get info including track timing etc. from FILE \
with `monk-ogg-probe-command'."
  (let (parse case-fold-search)
    (monk-ogg-probe file)		; `monk-ogg-info-str' is set
    (sleep-for monk-ogg-readin-wait)
    (cond
     ((equal monk-ogg-info-str -1) nil)
     (t
      (setq parse (monk-ogg-parse-info-str monk-ogg-info-str))
      (when parse
	(cons (cons (quote file) file)
	      parse))))))

;; Some traces of debugging.
;;
;; (defun monk-ogg-test-old (oggfile)
;;   (let ((monk-ogg-probe-command "/usr/local/bin/ogg123")
;; 	(monk-ogg-probe-options '("-vd" "null")))
;;     (monk-ogg-probe oggfile)))
;; 
;; (defun monk-ogg-test-new (oggfile)
;;   (let ((monk-ogg-probe-command "/usr/bin/ogg123")
;; 	(monk-ogg-probe-options '("-vd" "null")))
;;     (monk-ogg-probe oggfile)))
;;
;; (setq oggfile "/tmp/mp3/Bach - Toccata & Fugue - 02 - Prelude, Largo and Fugue in C: Prelude.ogg")
;; (setq oggfile (expand-file-name "~/.emacs"))
;; ogg123 from vorbis-tools 1.0rc3:
;; "Time: 00:00.09 [03:06.93] of 03:07.03  (112.5 kbps)  Output Buffer   3.1% "
;; Ogg123 from vorbis-tools 1.0beta4:
;; "Time: 00:00.01 [03:07.01] of 03:07.03, Bitrate: 301.4   "
;; (monk-ogg-test-new oggfile) "Time: 00:00.09 [03:06.93] of 03:07.03  (112.5 kbps)  Output Buffer   0.0% "
;; (monk-ogg-test-old oggfile) "Time: 00:00.01 [03:07.01] of 03:07.03, Bitrate: 301.4   "

;;
;; The track builder, still the same as of 0.5 in 0.6
;;

(defun monk-ogg-build-track-obj (file)
  "Get info from ogg file FILE, and build track object for the FILE.
`monk-ogg-get-info' is called."
  (let ((info (monk-ogg-get-info (expand-file-name file))))
    (when info
      (append (list
	       (cons 'player monk-ogg-player)
	       (cons 'track-opt nil)
	       (cons 'offset-opt monk-ogg-player-offset-opt)
	       (cons 'duration-opt monk-ogg-player-duration-opt)
	       (cons 'options monk-ogg-player-options)
	       (cons 'post-filter monk-ogg-player-post-filter))
	      info))))

(defun monk-ogg-insert-track-obj (file)
  "Insert track line at point."
  ;; (interactive "fogg File:")
  (let (obj fname frames time brate)
    (setq obj (monk-ogg-build-track-obj file)
	  frames (monk-assq-get-value 'frames obj)
	  time (monk-assq-get-value 'time obj))
    (cond ((and obj frames)
	   (setq
	    fname (file-name-nondirectory (monk-assq-get-value 'file obj))
	    ;; title (monk-assq-get-value 'title obj)
	    ;; artist (monk-assq-get-value 'artist obj)
	    brate (monk-assq-get-value 'bitrate obj))
	   (monk-insert-track
	    obj
	    "ogg:"
	    (format "%5s.00" (monk-sec2cdtime time))
	    (if brate
		(format "%3.1fk/s" brate)
	      "unknown")
	    frames
	    (format "\"%s\"" fname) 'newline)
	   obj)
	  (t (message "Track build faild for file: %s" file)))))

;; Should merge into monk.el as default of these variables later
;; (when (null (assq 'ogg monk-type-group-alist))
;;   (setq monk-track-generator-list
;; 	(cons 
;; 	 '("\\.ogg$" monk-ogg-insert-track-obj monk-ogg)
;; 	 monk-track-generator-list))
;;   (setq monk-emulate-loop-track-type (cons 'ogg
;; 					   monk-emulate-loop-track-type))
;;   (setq monk-type-group-alist
;; 	(cons '(ogg . dsp) monk-type-group-alist)))
