;;; monk-watch.el --- Watch ~/mp3/incoming, ~/ripped/<Artist> etc.

;; Started: Wed Jan 31 21:56:46 2001

;; This program is a sort of monk-specialized autorevert.el.
;; autorevert.el is not used here, so this works on emacs-19 as well.

;;
;; `monk-watch-directory-mode':     Minor mode starter / canceler.
;; `monk-watch-directory-handler':  Timer handler.
;; `monk-watch-check-directory':          Core of watcher.
;; `monk-watch-directory-interval': Defaults 5 (sec).
;;
;; The `monk-watch-directory-handler' has `monk-with-coding-system' brace,
;; and all the other subroutines called under the handler uses the coding
;; system settings.
;;

(require 'monk)
(provide 'monk-watch)

;; 
;; `monk-get-directory-files' -- "find . -type f" sort of.
;; Assumes `default-directory' is properly setup
;; Recurses itself -- barfs when tree is too deep.
;;

(defsubst monk-get-directory-files ()
  "Get a file name list under `default-directory' in full path description.
\"find . -type f\" compatible."
  ;; `default-directory' is assumed to be ending with slash.
  (let ((files (delete
		"." (delete
		     ".." (directory-files default-directory nil))))
	outl)
    (mapcar
     (lambda (f)
       (cond
	((file-accessible-directory-p f)
;;; mapcar() - cons() version.
;;;	   (mapcar (lambda (f)
;;;		     (setq outl (cons f outl)))
;;;		   (let ((default-directory
;;;			   (monk-put-trailing-/ (expand-file-name f))))
;;;		     (monk-get-directory-files)))
	 (setq outl (nconc
		     (nreverse
		      (let ((default-directory
			      (monk-put-trailing-/ (expand-file-name f))))
			(monk-get-directory-files))
		      ) outl)) )
	((file-readable-p f)
	 (setq outl (cons (concat default-directory f) outl)))
	(t nil)))
     files)
    (nreverse outl)))

;;
;; `monk-watch-check-directory' -- Update handler core.
;;

;;;###autoload
(defun monk-watch-check-directory (&optional justrevert)
  "Rebuild tracks if file associated with the track, and add new file \
lacked in buffer with `monk-find-file'.
Non-nil optional arg JUSTREVERT specifies just reverting existing tracks."
  (interactive "P")
  (or (consp (visited-file-modtime))	; failsafe
      (monk-time-stamp-this-buffer))
  (let ((dfiles (monk-get-directory-files))
	file
	size
	cfiles				; files in currently existing tracks
	nfiles)				; lacked (new) files in cfiles
    ;; Rebuilding,
    ;; (monk-walk-over-tracks ;; can't be used. must control `forward-line'
    (save-excursion
      (goto-char (point-min))
      (while (not (eobp))
	(setq file (monk-assq-get-value 'file (monk-get-track 'raw)))
	;;(message "X:%S Y:%S F:%S" nfiles cfiles file)
	;;(when file (message "FILE:%S" file))
	(cond
	 (file
	  (or justrevert (setq cfiles (cons file cfiles)))
	  (when
	      (monk-rebuild-track-if-newer-than-this-buffer file)
	    (forward-line 1)))
	 (t (forward-line 1))		; no 'file prop. leave it.
	 )))
    ;; Debugger
    ;; (message "nfiles:%S cfiles:%S file:%S" nfiles cfiles file)

    ;; Adding new files (only when (null justrevert))
    (when (null justrevert)
      (setq cfiles (nreverse cfiles))
      (mapcar (lambda (f)
		(or (member f cfiles)
		    (setq nfiles (cons f nfiles))
		    ))
	      dfiles)
      ;; (message "nfiles:%S cfiles:%S" nfiles cfiles)
      (when nfiles
	(setq nfiles (nreverse nfiles))
	(mapcar (lambda (f)
		  (when (and (file-readable-p f)
			     (setq size (nth 7 (file-attributes f)))
			     (and size (> size 0))
			     (monk-get-track-generator f))
		    (save-excursion
		      (goto-char (point-max))
		      (monk-find-file f))))
		nfiles)))
    (monk-time-stamp-this-buffer)))

;; Old `just-revert' version -- reverts existing tracks only.

;; (defun monk-revert-tracks ()
;;   "Revert existing tracks."
;;   (monk-only-with-monk-modes
;;    (monk-walk-over-tracks
;;     ;; (message "MonK doing %s" (monk-get-track 'raw)) ; DEBUGGER
;;     (monk-rebuild-track-if-newer-than-this-buffer))
;;    ;;(message "MonK done scan")		; DEBUGGER
;;    (monk-time-stamp-this-buffer nil)))	; stamp current time

;;
;; Subroutines concerning time comparison (buffer - file)
;;

;; Call when buffer is created, or by the starter,
;; *and* at the end of one cycle of `monk-revert-dir' finished.

(defun monk-time-stamp-this-buffer (&optional time)
  "Time stamp this buffer with optional arg TIME, \
or `current-time' output."
  (set-visited-file-modtime (or time (current-time))))

;; (visited-file-modtime) returns 0 (integer)
;; if buffer is not associated with a disk file.

(defun monk-file-newer-than-this-buffer-p (file)
  "Test if FILE is newer than this buffer.
Returns FILE's modification time if true, else nil."
  (or (consp (visited-file-modtime))
      (error "Buggy call for monk-file-newer-than-current-buffer-p."))
  (let* ((bta (car (visited-file-modtime)))
	 (btb (cdr (visited-file-modtime)))
	 (ft (nth 5 (file-attributes file)))
	 (fta (car ft))
	 (ftb (cadr ft)))
    (cond
     ((> fta bta) ft)
     ((and (= fta bta) (> ftb btb)) ft)
     (t nil))))

(defun monk-rebuild-track-if-newer-than-this-buffer (&optional file)
  "Call `monk-rebuild-track' if file associated with current track is \
newer than this buffer.
If optional arg FILE is supplied the FILE name is used
instead of 'file property of the current track.
Returns nil if the track is removed, else t."
  (let ((file 
	 (or file (monk-assq-get-value 'file (monk-get-track 'raw)))))
    (cond
     ((and file (null (file-readable-p file)))
      ;; (monk-rebuild-track 0) ;; Do it from lower level:
      (monk-remove-current-track 1)
      nil)
     ((and file (monk-file-newer-than-this-buffer-p file))
      (monk-rebuild-track)
      t)
     (t t))))				; nop. it's a good track.

;;
;; Defining `monk-watch-directory-mode' as a minor mode.
;;

(defvar monk-watch-directory-mode nil
  "A Minor mode. Should be set only thru `monk-watch-directory-mode' command.
Internal use only. Use command `monk-watch-directory-mode' to control this.")
(make-variable-buffer-local 'monk-watch-directory-mode)

(defvar monk-watch-directory-interval 6 ; prime each other to autorevert.el's
  "*Interval for `monk-watch-directory-mode'.
Currently, consistency of this value is not checked.
If too many files are being updated, you may increase this interval, or
emacs might become too busy.")

(defvar monk-watch-directory-add-new-file t
  "*Defaults t, which enables adding new file\(s\) in the `default-directory'
in the minor mode `monk-watch-directory-mode'.
Referred by `monk-watch-check-directory'.")
(make-variable-buffer-local 'monk-watch-directory-mode)

(defvar monk-watch-directory-mode-buffers nil
  "List of buffers in `monk-watch-directory-mode'. Internal use only.")

(defvar monk-watch-directory-timer-obj nil "Timer object. Internal use only.")

(defsubst monk-turn-off-watch-directory-timer ()
  "Cancel timer for `monk-watch-directory-mode'."
  (when monk-watch-directory-timer-obj
    (cancel-timer monk-watch-directory-timer-obj)))

(defsubst monk-fire-up-watch-directory-timer ()
  "Fire up timer for handler function `monk-watch-directory-handler'."
  (setq monk-watch-directory-timer-obj
	(run-at-time (format "%ssec" monk-watch-directory-interval)
		     monk-watch-directory-interval
		     'monk-watch-directory-handler)))

(add-hook 'kill-buffer-hook (lambda () (monk-watch-directory-mode -1)))

(defsubst monk-watch-add-buffer ()
  (setq monk-watch-directory-mode-buffers
	(cons (current-buffer) monk-watch-directory-mode-buffers)))

(defsubst monk-watch-del-buffer ()
  (setq monk-watch-directory-mode-buffers
	(delq (current-buffer) monk-watch-directory-mode-buffers)))

;;;###autoload
(defun monk-watch-directory-mode (arg)
  "Ordinary minor mode controller command for `monk-watch-directory-mode'.
Toggles `monk-watch-directory-mode'.if ARG is nil, or
sets `monk-watch-directory-mode' to t if ARG is greater than 0."
  (interactive "P")
  (monk-only-with-monk-modes ;; progn ;;
   (setq monk-watch-directory-mode
	 (cond ((null arg)
		(not monk-watch-directory-mode))
	       (t (> (prefix-numeric-value arg) 0))))
   (cond
    (monk-watch-directory-mode		; ENABLED
     (monk-time-stamp-this-buffer)
     (cond
      (monk-watch-directory-mode-buffers
       (monk-watch-add-buffer))
      (t			   ; this is the first buffer to watch
       (monk-watch-add-buffer)
       (monk-fire-up-watch-directory-timer))))
    (t					; DISABLED
     (monk-watch-del-buffer)
     (when (null monk-watch-directory-mode-buffers)
       (monk-turn-off-watch-directory-timer)) ))
   (message "monk-watch-directory-mode is turned %s"
	    (if monk-watch-directory-mode "on" "off"))
   monk-watch-directory-mode))

(defun monk-watch-directory-handler ()
  "The handler function for `monk-watch-directory-mode'."
  (monk-with-coding-system
   (mapcar (lambda (b)
	     (or (input-pending-p)
		 (monk-with-current-buffer
		  b
		  (let ((monk-rebuild-track-auto nil))
		    (monk-watch-check-directory
		     (null monk-watch-directory-add-new-file))))))
	   monk-watch-directory-mode-buffers)))

(defun monk-toggle-watch-directory-add-new-file (&optional arg)
  "Toggles `monk-watch-directory-add-new-file' status.
Optional non-nil ARG or numeric ARG >0 forces turns it on,
or nil ARG or numeric ARG <=0 turns it off."
  (interactive "P")
  (message "`monk-watch-directory-add-new-file' set to %S"
	   (setq monk-watch-directory-add-new-file
		 (cond
		  ((null arg) (not monk-watch-directory-add-new-file))
		  ((consp arg) t)
		  ((> (prefix-numeric-value current-prefix-arg) 0) t)
		  ((<= (prefix-numeric-value current-prefix-arg) 0) nil)
		  (t arg)))))		; non-interactive call ?

;; Registering the minor-mode

(or (assq 'monk-watch-directory-mode minor-mode-alist)
    (setq minor-mode-alist
	  (cons (list 'monk-watch-directory-mode " DirWatch")
		minor-mode-alist)))
