;;; monk-wav.el --- .wav track generator for MonK

;; (c) 2001,2002,2003 Seiichi Namba <sn@asahi-net.email.ne.jp>
;; Started: Wed Jun  6 12:05:00 2001 s.n.

;; BUGS: 
;;    (1) monk-wav script can play .aiff etc., but can probe .wav only.
;;    (2) 'fps is assumed to be 1000. I don't know if this is correct.
;;    (3) Your sox(1) is assumed to be able to output to /dev/dsp.

;; The track format:
;; ((file . "/tmp/wav/o.wav")
;;  (player . "monk-wav")
;;  (track-opt)
;;  (offset-opt . "-o %d")
;;  (duration-opt . "-d %d")
;;  (options)
;;  (post-filter)
;;  (frames . 31000)
;;  (time . 31)
;;  (cent . 310)
;;  (fps . 1000)
;;  (type . wav))

(require 'monk)
(require 'dired)
(require 'dired-aux)
(provide 'monk-wav)

(defconst monk-wav-version "0.3")

(defconst monk-wav-player "monk-wav")

(defconst monk-wav-probe-command "monk-wav")
(defconst monk-wav-probe-options '("-p"))

(defvar monk-wav-player-options nil "*For `monk-wav-player'")

(defvar monk-wav-player-offset-opt "-o %s")
(defvar monk-wav-player-duration-opt "-d %s")
(defvar monk-wav-player-post-filter nil)

(defvar monk-wav-readin-wait 0
  "A delay in second inserted after reading .wav file.")

(defun monk-wav-get-info (file)
  "Get info including track timing etc. from FILE \
with `monk-wav-probe-command'."
  (let ((buf (get-buffer-create monk-info-buffer)) tem)
    (save-excursion
      (set-buffer buf) (erase-buffer)
      (setq default-directory
	    (monk-put-trailing-/ (file-name-directory file)))
      (cond
       (
	(= 0 (setq tem ;; hack `tem' if debug required.
		   (apply 'call-process monk-wav-probe-command nil t nil
			  (append monk-wav-probe-options
				  (list
				   file ; expanded already
				   )))))
	;; Delay. Recursive directory read sometimes failed (?)
	;; (sleep-for (/ monk-wait-proc-to-die 2))
	(sleep-for monk-wav-readin-wait)
	(cons (cons (quote file) file)
	      (monk-wav-parse-info buf)))))))

(defun monk-wav-parse-info (buf)
  "Parse buffer BUF and generate 'time, 'frames, 'cent, 'fps, and 'type."
  (save-excursion
    (set-buffer buf)
    (let* ((time
	    (string-to-number
	     (car (delete
		   "" (dired-split
		       "\n" (buffer-substring-no-properties
			     (point-min) (point-max)))))))
	   (frames (* time 1000)))
      (list
       (cons 'time time)
       (cons 'frames frames)
       (cons 'cent (/ frames 100.0))
       (cons 'fps 1000)
       (cons 'type 'wav)
       ;; (cons 'brate (buffer-substring (match-beginning 4) (match-end 4)))
       ))))

(defun monk-wav-build-track-obj (file)
  "Get info from wav file FILE, and build track object for the FILE.
`monk-wav-get-info' is called."
  (let ((info (monk-wav-get-info (expand-file-name file))))
    (when info
      (append (list
	       (cons 'player monk-wav-player)
	       (cons 'track-opt nil)
	       (cons 'offset-opt monk-wav-player-offset-opt)
	       (cons 'duration-opt monk-wav-player-duration-opt)
	       (cons 'options monk-wav-player-options)
	       (cons 'post-filter monk-wav-player-post-filter))
	      info))))

(defun monk-wav-insert-track-obj (file)
  "Insert track line at point."
  ;; (interactive "fwav File:")
  (let (obj fname frames time)
    (setq obj (monk-wav-build-track-obj file)
	  frames (monk-assq-get-value 'frames obj)
	  time (monk-assq-get-value 'time obj))
    (cond ((and obj frames)
	   (setq
	    fname (file-name-nondirectory (monk-assq-get-value 'file obj))
	    ;; title (monk-assq-get-value 'title obj)
	    ;; artist (monk-assq-get-value 'artist obj)
	    )
	   (monk-insert-track
	    obj
	    "wav:"
	    (format "%5s.00" (monk-sec2cdtime time))
	    0
	    frames
	    (format "\"%s\"" fname) 'newline)
	   obj)
	  (t (message "Track build faild for file: %s" file)))))

;; Should merge into monk.el as default of these variables later
;; (when (null (assq 'wav monk-type-group-alist))
;;   (setq monk-track-generator-list
;; 	(cons 
;; 	 '("\\.wav$" monk-wav-insert-track-obj monk-wav)
;; 	 monk-track-generator-list))
;;   ;; Not required. sox(1) has "trim start length" option.
;;   ;;(setq monk-emulate-loop-track-type (cons 'wav monk-emulate-loop-track-type))
;;   (setq monk-type-group-alist
;; 	(cons '(wav . dsp) monk-type-group-alist)))

;; Probe exsample for future versions.

;; $ time sox -V o.wav -t .wav /dev/null trim 0 1       
;; sox: Reading Wave file: Microsoft PCM format, 2 channels, 44100 samp/sec
;; sox:         176400 byte/sec, 4 block align, 16 bits/samp, 5582848 data bytes
;; sox: Input file: using sample rate 44100
;;         size shorts, encoding signed (2's complement), 2 channels
;; sox: Input file: comment "o.wav"
;; 
;; sox: Writing Wave file: Microsoft PCM format, 2 channels, 44100 samp/sec
;; sox:         176400 byte/sec, 4 block align, 16 bits/samp
;; sox: Output file: using sample rate 44100
;;         size shorts, encoding signed (2's complement), 2 channels
;; sox: Output file: comment "o.wav"
;; 
;; passed done
;; sox: Finished writing Wave file, 176 data bytes 88 samples
;; 
;; sox -V o.wav -t .wav /dev/null trim 0 1  0.44s user 0.05s system 99% cpu 0.492 total

;; $ time sox -V o.aiff -t .wav /dev/null trim 0 1      
;; Comment:     "o.wav "
;; Unity MIDI Note: 0
;; Low   MIDI Note: 0
;; High  MIDI Note: 0
;; sox: Input file: using sample rate 44100
;;         size shorts, encoding signed (2's complement), 2 channels
;; sox: Input file: comment "o.wav "
;; 
;; sox: Writing Wave file: Microsoft PCM format, 2 channels, 44100 samp/sec
;; sox:         176400 byte/sec, 4 block align, 16 bits/samp
;; sox: Output file: using sample rate 44100
;;         size shorts, encoding signed (2's complement), 2 channels
;; sox: Output file: comment "o.wav "
;; 
;; passed done
;; sox: Finished writing Wave file, 176 data bytes 88 samples
;; 
;; sox -V o.aiff -t .wav /dev/null trim 0 1  0.51s user 0.08s system 100% cpu 0.589 total
